import logging
from typing import Any, Callable, Dict, Generic, Optional, Set, TypeVar

from local_tuya.device import State
from local_tuya.domoticz.types import DomoticzUnit
from local_tuya.domoticz.units.base import Unit, UnitCommand
from local_tuya.domoticz.units.ids import UnitId

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=State)
V = TypeVar("V")


class UnitManager(Generic[T]):
    def __init__(
        self,
        name: str,
        units: Dict[int, DomoticzUnit],
        included_units: Optional[Set[UnitId]],
    ):
        super().__init__()
        self._name = name
        self._domoticz_units = units
        self._included_units = included_units
        self._units: Dict[int, Unit] = {}
        self._value_from_state: Dict[int, Callable[[T], Any]] = {}

    def register(self, unit: Unit[V], value_from_state: Callable[[T], V]) -> None:
        if not self._included_units or unit.id in self._included_units:
            self._units[unit.id] = unit
            self._value_from_state[unit.id] = value_from_state
            unit.ensure(self._domoticz_units.get(unit.id), self._name)
        elif unit.id in self._domoticz_units:
            self._domoticz_units[unit.id].Delete()

    async def on_command(self, unit_id: int, command: UnitCommand) -> None:
        unit = self._units.get(unit_id)
        if unit:
            logger.debug("sending command %s to unit %s", command, unit_id)
            await unit.on_command(command)

    def update(self, state: T) -> None:
        logger.debug("received new device state: %s", state)
        for id_, unit in self._units.items():
            unit.update(self._value_from_state[id_](state))

    def cleanup_domoticz_references(self) -> None:
        """Prevent keeping references to domoticz units."""
        self._domoticz_units = {}
        self._units = {}



import asyncio
import logging

from concurrent_tasks import BackgroundTask

from local_tuya.errors import CommandTimeoutError
from local_tuya.events import EventNotifier
from local_tuya.protocol.events import (
    CommandSent,
    ConnectionEstablished,
    ConnectionLost,
)
from local_tuya.protocol.message import HeartbeatCommand

logger = logging.getLogger(__name__)


class Heartbeat(BackgroundTask):
    def __init__(self, interval: float, event_notifier: EventNotifier):
        super().__init__(self._run, interval, event_notifier)
        event_notifier.register(ConnectionLost, lambda _: self.cancel())
        event_notifier.register(ConnectionEstablished, lambda _: self.create())

    def __enter__(self) -> "Heartbeat":
        return self

    async def _run(self, interval: float, event_notifier: EventNotifier) -> None:
        while True:
            try:
                await event_notifier.emit(CommandSent(HeartbeatCommand()))
            except CommandTimeoutError:
                logger.warning("timeout waiting for heartbeat response")
            await asyncio.sleep(interval)



from local_tuya.protocol.message.handlers import MessageHandler, get_handler
from local_tuya.protocol.message.messages import (
    Command,
    HeartbeatCommand,
    HeartbeatResponse,
    Response,
    StateCommand,
    StateResponse,
    StatusResponse,
    UpdateCommand,
    UpdateResponse,
    Value,
    Values,
)



from dataclasses import dataclass
from typing import ClassVar

from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad, unpad


@dataclass
class AESCipher:
    key: bytes

    BLOCK_SIZE: ClassVar[int] = 16
    MODE: ClassVar[int] = AES.MODE_ECB

    def encrypt(self, s: bytes) -> bytes:
        if not s:
            return s
        cipher = AES.new(self.key, mode=self.MODE)  # type: ignore
        return cipher.encrypt(pad(s, self.BLOCK_SIZE))

    def decrypt(self, s: bytes) -> bytes:
        if not s:
            return s
        if len(s) % self.BLOCK_SIZE:
            raise ValueError(
                f"length {len(s)} should be a multiple of {self.BLOCK_SIZE}"
            )
        cipher = AES.new(self.key, self.MODE)  # type: ignore
        return unpad(cipher.decrypt(s), self.BLOCK_SIZE)



import asyncio
import logging
from typing import Awaitable, Callable, Optional, cast

from local_tuya.protocol import Values

logger = logging.getLogger(__name__)


class StateHandler:
    def __init__(self, updated_callback: Callable[[Values], Awaitable]):
        self._state: Optional[Values] = None
        self._updated_event = asyncio.Event()
        self._updated_callback = updated_callback

    async def updated(self, state: Values) -> None:
        self._state = state
        self._updated_event.set()
        await self._updated_callback(state)

    async def state(self) -> Values:
        if self._state is None:
            await self._updated_event.wait()
        return cast(Values, self._state).copy()

    async def matches(self, values: Values) -> None:
        """Wait until the device state matches the update values."""
        if not self._state:
            await self._updated_event.wait()

        def _matches() -> bool:
            for k, v in values.items():
                if cast(Values, self._state).get(k) != v:
                    return False
            return True

        while not _matches():
            self._updated_event.clear()
            logger.debug("waiting for device state to be updated")
            await self._updated_event.wait()



from typing import Dict, Optional, Set, Tuple

from local_tuya.device.enums import DataPoint
from local_tuya.protocol import Value, Values


class Constraint:
    """The data point is constrained if any of the exclusions is active."""

    def __init__(
        self,
        data_point: DataPoint,
        # An exclusion is active if its datapoint is in the set of values.
        *exclusions: Tuple[DataPoint, Set[Value]],
        # Only check the constraint if the data point is in those values.
        restrict_to: Optional[Set[Value]] = None,
    ):
        self.data_point = data_point
        self._exclusions: Dict[str, Set[Value]] = {
            data_point: values for data_point, values in exclusions
        }
        self._restrict_to = restrict_to

    def applicable(self, value: Value) -> bool:
        """Return whether the constraint should be checked."""
        return self._restrict_to is None or value in self._restrict_to

    def ko(self, values: Values) -> bool:
        """Returns whether an exclusion applies on the given values."""
        return any(
            v in self._exclusions[k] for k, v in values.items() if k in self._exclusions
        )


class Constraints:
    """Set of constraints for a device that forbids certain commands."""

    def __init__(self, *constraints: Constraint):
        self._constraints: Dict[str, Constraint] = {
            constraint.data_point: constraint for constraint in constraints
        }

    def filter_values(self, values: Values, current: Values) -> Values:
        """Filter values that can be updated given the device constraints."""
        # Check on merged values.
        merged = {**current, **values}
        filtered: Values = {}
        for data_point, value in values.items():
            constraint = self._constraints.get(data_point)
            if constraint and constraint.applicable(value) and constraint.ko(merged):
                continue
            filtered[data_point] = value
        return filtered



import logging

from local_tuya.events import EventNotifier
from local_tuya.protocol.events import ConnectionLost, DataReceived, ResponseReceived
from local_tuya.protocol.message import MessageHandler

logger = logging.getLogger(__name__)


class Receiver:
    def __init__(
        self,
        message_handler: MessageHandler,
        event_notifier: EventNotifier,
    ):
        event_notifier.register(DataReceived, self._receive_messages)
        event_notifier.register(ConnectionLost, self._reset)
        self._handler = message_handler
        self._notifier = event_notifier
        self._buffer = b""

    def _reset(self, _: ConnectionLost) -> None:
        self._buffer = b""

    async def _receive_messages(self, data: DataReceived) -> None:
        self._buffer += data
        try:
            # Get all message from the data.
            while self._buffer:
                (
                    sequence_number,
                    response,
                    command_class,
                    self._buffer,
                ) = self._handler.unpack(self._buffer)
                if not response:
                    # We need more data to build a response.
                    break
                logger.debug(
                    "received message %i %s",
                    sequence_number,
                    response.__class__.__name__,
                )
                await self._notifier.emit(
                    ResponseReceived(
                        sequence_number,
                        response,
                        command_class,
                    )
                )
        except Exception:
            # Reset the buffer.
            self._buffer = b""



from dataclasses import dataclass
from enum import Enum

from local_tuya.backoff import Backoff, SequenceBackoff


class Version(bytes, Enum):
    """Supported versions."""

    v33 = b"3.3"


@dataclass
class ProtocolConfig:
    id_: str
    address: str
    key: bytes
    port: int = 6668
    version: Version = Version.v33
    # How long to wait between reconnection attempts.
    connection_backoff: Backoff = SequenceBackoff(1, 5, 10, 30, 60, 300)
    # Seconds to wait until command can be confirmed.
    # This excludes time waiting for the connection to be established.
    timeout: float = 5
    # Seconds between each heartbeat interval keeping the connection alive.
    heartbeat_interval: float = 15
    # How long to keep the state until a refresh is done.
    # State is maintained via status updates so a low value shouldn't be required.
    state_refresh_interval: float = 3600
